import concurrent.futures
import os
import sys
import json
from reademptionlib.bammerger import BamMerger
from reademptionlib.coveragecalculator import CoverageCalculator
from reademptionlib.crossalignfilter import CrossAlignFilter
from reademptionlib.deseq import DESeqRunner
from reademptionlib.fasta import FastaParser
from reademptionlib.genewisequanti import GeneWiseQuantification
from reademptionlib.genewisequanti import GeneWiseOverview
from reademptionlib.paths import Paths
from reademptionlib.projectcreator import ProjectCreator
from reademptionlib.rawstatdata import RawStatDataWriter, RawStatDataReader
from reademptionlib.readaligner import ReadAligner
from reademptionlib.readalignerstats import ReadAlignerStats
from reademptionlib.readrealigner import ReadRealigner
from reademptionlib.readalignerstatstable import ReadAlignerStatsTable
from reademptionlib.readprocessor import ReadProcessor
from reademptionlib.sambamconverter import SamToBamConverter
from reademptionlib.wiggle import WiggleWriter


class Controller(object):

    """Manage the actions of the subcommands.

    The Controller take care of providing the argumentes like path
    names and the parallel processing of tasks.

    """

    def __init__(self, args):
        """Create an instance."""
        self._args = args
        self._paths = Paths(args.project_path)
        self._read_files = None
        self._ref_seq_files = None

    def create_project(self, version):
        """Create a new project."""
        sys.stdout.write(
            "   ___  _______   ___                 __  _\n"
            "  / _ \/ __/ _ | / _ \___ __ _  ___  / /_(_)__  ___\n"
            " / , _/ _// __ |/ // / -_)  ' \/ _ \/ __/ / _ \/ _ \\\n"
            "/_/|_/___/_/ |_/____/\__/_/_/_/ .__/\__/_/\___/_//_/\n"
            "                             / /\n"
            "====================================================\n"
            "========================================\n"
            "=======================\n"
            "==============\n\n"
            "[https://reademption.readthedocs.io/en/latest/]\n\n")
        project_creator = ProjectCreator()
        project_creator.create_root_folder(self._args.project_path)
        project_creator.create_subfolders(self._paths.required_folders())
        project_creator.create_version_file(self._paths.version_path, version)
        sys.stdout.write("Created folder \"%s\" and required subfolders.\n" % (
            self._args.project_path))
        sys.stdout.write("Please copy read files into folder \"%s\" and "
                         "reference sequences files into folder \"%s\".\n" % (
                             self._paths.read_fasta_folder,
                             self._paths.ref_seq_folder))

    def align_reads(self):
        """Perform the alignment of the reads."""
        self._test_folder_existance(
            self._paths.required_read_alignment_folders())
        assert self._args.paired_end in [True, False]
        self._ref_seq_files = self._paths.get_ref_seq_files()
        self._paths.set_ref_seq_paths(self._ref_seq_files)
        self._test_align_file_existance()
        if not self._args.paired_end:
            # Single end reads
            self._read_files = self._paths.get_read_files()
            self._lib_names = self._paths.get_lib_names_single_end()
            self._paths.set_read_files_dep_file_lists_single_end(
                self._read_files, self._lib_names)
            if not self._args.realign:
                self._set_primary_aligner_paths_to_final_paths()
            self._prepare_reads_single_end()
            self._align_single_end_reads()
        else:
            # Paired end reads
            self._read_file_pairs = self._paths.get_read_file_pairs()
            self._lib_names = self._paths.get_lib_names_paired_end()
            self._paths.set_read_files_dep_file_lists_paired_end(
                self._read_file_pairs, self._lib_names)
            if not self._args.realign:
                self._set_primary_aligner_paths_to_final_paths()
            self._prepare_reads_paired_end()
            self._align_paired_end_reads()
        #self._sam_to_bam(
        #    self._paths.primary_read_aligner_sam_paths,
        #    self._paths.primary_read_aligner_bam_prefix_paths,
        #    self._paths.primary_read_aligner_bam_paths)
        self._generate_read_alignment_stats(
            self._lib_names,
            self._paths.primary_read_aligner_bam_paths,
            self._paths.unaligned_reads_paths,
            self._paths.primary_read_aligner_stats_path)
        final_unaligned_reads_paths = self._paths.unaligned_reads_paths
        if self._args.realign:
            self._run_realigner_and_process_alignments()
            self._merge_bam_files()
            final_unaligned_reads_paths = (
                self._paths.realigned_unaligned_reads_paths)
        if self._args.crossalign_cleaning_str is not None:
            self._remove_crossaligned_reads()
        self._generate_read_alignment_stats(
            self._lib_names,
            self._paths.read_alignment_bam_paths,
            final_unaligned_reads_paths,
            self._paths.read_alignments_stats_path)
        self._write_alignment_stat_table()

    def _remove_crossaligned_reads(self):
        self._string_to_species_and_sequence_ids()
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for (bam_path, bam_with_crossmappings_path,
                 bam_cleaned_tmp_path, crossmapped_reads_path) in zip(
                    self._paths.read_alignment_bam_paths,
                    self._paths.read_alignment_bam_with_crossmappings_paths,
                    self._paths.read_alignment_bam_cross_cleaned_tmp_paths,
                    self._paths.crossmapped_reads_paths):
                jobs.append(executor.submit(
                    self._remove_crossaligned_reads_for_lib, bam_path,
                    bam_with_crossmappings_path, bam_cleaned_tmp_path,
                    crossmapped_reads_path))
        # Evaluate thread outcome
        self._check_job_completeness(jobs)

    def _remove_crossaligned_reads_for_lib(
            self, bam_path, bam_with_crossmappings_path, bam_cleaned_tmp_path,
            crossmapped_reads_path):
        # Perform the removal or cross aligned reads
        cross_align_filter = CrossAlignFilter(
            bam_path, bam_cleaned_tmp_path, crossmapped_reads_path,
            self._species_and_sequence_ids)
        cross_align_filter.determine_crossmapped_reads()
        cross_align_filter.write_crossmapping_free_bam()
        # Rename the original mapping file that potentially
        # contains cross aligned reads
        os.rename(bam_path, bam_with_crossmappings_path)
        os.rename(bam_path + ".bai", bam_with_crossmappings_path + ".bai")
        # Move the cross aligned filtered file to the final mapping
        # path
        os.rename(bam_cleaned_tmp_path, bam_path)
        os.rename(bam_cleaned_tmp_path + ".bai", bam_path + ".bai")

    def _string_to_species_and_sequence_ids(self):
        self._species_and_sequence_ids = {}
        orgs_and_seq_ids_strs = self._args.crossalign_cleaning_str.split(";")
        if len(orgs_and_seq_ids_strs) < 2:
            self._write_err_msg_and_quit(
                "Error! Only one organism is defined for the cross align "
                "removal. This does not make sense.\nYou gave the "
                "following input:\n%s\n" % self._args.crossalign_cleaning_str)
        for org_and_seq_ids_str in orgs_and_seq_ids_strs:
            org, seq_ids_str = org_and_seq_ids_str.strip().split(":")
            seq_ids = [seq_id.strip() for seq_id in seq_ids_str.split(",")]
            if "" in seq_ids:
                seq_ids.remove("")
            if len(seq_ids) < 1:
                self._write_err_msg_and_quit(
                    "Error! No sequence ID was given for the species '%s'. "
                    "This does not make sense.\nYou gave the "
                    "following input:\n%s\n" % (
                        org, self._args.crossalign_cleaning_str))
            self._species_and_sequence_ids[org] = seq_ids

    def _set_primary_aligner_paths_to_final_paths(self):
        # If no remapping is performed the paths of the final bam files
        # is the paths of the primary mapper
        self._paths.primary_read_aligner_bam_prefix_paths = (
            self._paths.read_alignment_bam_prefix_paths)
        self._paths.primary_read_aligner_bam_paths = (
            self._paths.read_alignment_bam_paths)
        self._paths.primary_read_aligner_stats_path = (
            self._paths.read_alignments_stats_path)

    def _merge_bam_files(self):
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for merged_bam, primary_aligner_bam, realigner_bam in zip(
                    self._paths.read_alignment_bam_paths,
                    self._paths.primary_read_aligner_bam_paths,
                    self._paths.read_realigner_bam_paths):
                bam_merger = BamMerger()
                jobs.append(executor.submit(
                    bam_merger.merge, merged_bam,
                    primary_aligner_bam, realigner_bam))
        self._check_job_completeness(jobs)
        if not self._args.keep_original_alignments:
            for bam_file_list in [
                    self._paths.primary_read_aligner_bam_paths,
                    self._paths.read_realigner_bam_paths]:
                for bam_file in bam_file_list:
                    os.remove(bam_file)
                    os.remove("%s.bai" % bam_file)

    def _run_realigner_and_process_alignments(self):
        # As the realigner needs a *sorted* SAM file
        self._generate_sorted_tmp_sam_file()
        self._realign_unmapped_reads()
        #self._sam_to_bam(
        #    self._paths.read_realigner_sam_paths,
        #    self._paths.read_realigner_bam_prefixes_paths,
        #    self._paths.read_realigner_sam_paths)
        self._generate_read_alignment_stats(
            self._lib_names,
            self._paths.read_realigner_bam_paths,
            self._paths.realigned_unaligned_reads_paths,
            self._paths.read_realigner_stats_path)

    def _test_align_file_existance(self):
        """Test if the input file for the the align subcommand exist."""
        if len(self._paths.get_read_files()) == 0:
            self._write_err_msg_and_quit("Error! No read libraries given!\n")
        if len(self._ref_seq_files) == 0:
            self._write_err_msg_and_quit(
                "Error! No reference sequence files given!\n")

    def _test_folder_existance(self, task_specific_folders):
        """Test the existance of required folders."""
        for folder in (
                self._paths.required_base_folders() + task_specific_folders):
            if not os.path.exists(folder):
                self._write_err_msg_and_quit(
                    "Error! Folder '%s' does not exist! Is the given project "
                    "folder name correct?\n" % folder)

    def _file_needs_to_be_created(self, file_path, quiet=False):
        """Test if a file exists of need to be created."""
        if not self._args.check_for_existing_files:
            return True
        if os.path.exists(file_path):
            if not quiet:
                sys.stderr.write(
                    "File %s exists. Skipping its generation.\n" % file_path)
            return False
        return True

    def _prepare_reads_single_end(self):
        """Manage the prepartion of reads before the actual mappings."""
        read_files_and_jobs = {}
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for lib_name, read_path, processed_read_path in zip(
                    self._lib_names, self._paths.read_paths,
                    self._paths.processed_read_paths):
                if not self._file_needs_to_be_created(
                        processed_read_path):
                    continue
                read_processor = ReadProcessor(
                    poly_a_clipping=self._args.poly_a_clipping,
                    min_read_length=self._args.min_read_length,
                    fastq=self._args.fastq,
                    min_phred_score=self._args.min_phred_score,
                    adapter=self._args.adapter,
                    reverse_complement=self._args.reverse_complement)
                read_files_and_jobs[lib_name] = executor.submit(
                    read_processor.process_single_end, read_path,
                    processed_read_path)
        self._evaluet_job_and_generate_stat_file(read_files_and_jobs)

    def _prepare_reads_paired_end(self):
        read_files_and_jobs = {}
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for lib_name, read_path_pair, processed_read_path_pair in zip(
                self._lib_names, self._paths.read_path_pairs,
                    self._paths.processed_read_path_pairs):
                for processed_read_path in processed_read_path_pair:
                    if not self._file_needs_to_be_created(
                            processed_read_path):
                        continue
                    read_processor = ReadProcessor(
                        poly_a_clipping=False,
                        min_read_length=self._args.min_read_length,
                        fastq=self._args.fastq,
                        min_phred_score=self._args.min_phred_score,
                        adapter=self._args.adapter)
                    read_files_and_jobs[lib_name] = executor.submit(
                        read_processor.process_paired_end, read_path_pair,
                        processed_read_path_pair)
        self._evaluet_job_and_generate_stat_file(read_files_and_jobs)

    def _evaluet_job_and_generate_stat_file(self, read_files_and_jobs):
        raw_stat_data_writer = RawStatDataWriter(pretty=True)
        # Evaluate thread outcome
        self._check_job_completeness(read_files_and_jobs.values())
        if not self._file_needs_to_be_created(
                self._paths.read_processing_stats_path):
            return
        # Create a dict of the read file names and the processing
        # counting results
        read_files_and_stats = dict(
            [(lib_name, job.result()) for lib_name, job in
             read_files_and_jobs.items()])
        raw_stat_data_writer.write(
            read_files_and_stats, self._paths.read_processing_stats_path)

    def _align_single_end_reads(self):
        """Manage the actual alignment of single end reads."""
        read_aligner = ReadAligner(
            self._args.segemehl_bin, self._args.progress)
        if self._file_needs_to_be_created(self._paths.index_path):
            read_aligner.build_index(
                self._paths.ref_seq_paths, self._paths.index_path)
        for read_path, output_path, nomatch_path, bam_path in zip(
            self._paths.processed_read_paths,
            self._paths.primary_read_aligner_bam_paths,
            self._paths.unaligned_reads_paths,
                self._paths.read_alignment_bam_paths):
            if not self._file_needs_to_be_created(output_path):
                continue
            elif not self._file_needs_to_be_created(bam_path):
                continue
            read_aligner.run_alignment(
                read_path, self._paths.index_path, self._paths.ref_seq_paths,
                output_path, nomatch_path, int(self._args.processes),
                int(self._args.segemehl_accuracy),
                float(self._args.segemehl_evalue), self._args.split,
                paired_end=False)

    def _align_paired_end_reads(self):
        """Manage the actual alignemnt of paired end reads."""
        read_aligner = ReadAligner(
            self._args.segemehl_bin, self._args.progress)
        if self._file_needs_to_be_created(self._paths.index_path):
            read_aligner.build_index(
                self._paths.ref_seq_paths, self._paths.index_path)
        for read_path_pair, output_path, nomatch_path, bam_path in zip(
            self._paths.processed_read_path_pairs,
            self._paths.primary_read_aligner_sam_paths,
            self._paths.unaligned_reads_paths,
                self._paths.primary_read_aligner_bam_paths):
            if not self._file_needs_to_be_created(output_path):
                continue
            elif not self._file_needs_to_be_created(bam_path):
                continue
            read_aligner.run_alignment(
                read_path_pair, self._paths.index_path,
                self._paths.ref_seq_paths, output_path, nomatch_path,
                int(self._args.processes), int(self._args.segemehl_accuracy),
                float(self._args.segemehl_evalue), self._args.split,
                paired_end=True)

    def _realign_unmapped_reads(self):
        read_realigner = ReadRealigner(
            self._args.lack_bin, self._args.progress)
        for (query_fasta_path, query_sam_path, realignment_sam_path,
             unaligned_reads_path) in zip(
                self._paths.unaligned_reads_paths,
                self._paths.read_realigner_tmp_sam_paths,
                self._paths.read_realigner_sam_paths,
                self._paths.realigned_unaligned_reads_paths):
            read_realigner.run_alignment(
                query_fasta_path, query_sam_path, self._paths.ref_seq_paths,
                realignment_sam_path, unaligned_reads_path,
                int(self._args.processes), int(self._args.segemehl_accuracy))
            os.remove(query_sam_path)

    def _sam_to_bam(self, sam_paths, bam_prefixes_paths, bam_paths):
        """Manage the conversion of mapped read from SAM to BAM format."""
        sam_to_bam_converter = SamToBamConverter()
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for sam_path, bam_prefix_path, bam_path in zip(
                    sam_paths, bam_prefixes_paths, bam_paths):
                if not self._file_needs_to_be_created(bam_path):
                    continue
                jobs.append(executor.submit(
                    sam_to_bam_converter.sam_to_bam,
                    sam_path, bam_prefix_path))
        # Evaluate thread outcome
        self._check_job_completeness(jobs)

    def _generate_sorted_tmp_sam_file(self):
        sam_to_bam_converter = SamToBamConverter()
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for bam_path, sam_path in zip(
                    self._paths.primary_read_aligner_bam_paths,
                    self._paths.read_realigner_tmp_sam_paths):
                jobs.append(executor.submit(
                    sam_to_bam_converter.bam_to_sam, bam_path, sam_path))
        # Evaluate thread outcome
        self._check_job_completeness(jobs)

    def _generate_read_alignment_stats(
            self, lib_names, result_bam_paths, unaligned_reads_paths,
            output_stats_path):
        """Manage the generation of alingment statistics."""
        raw_stat_data_writer = RawStatDataWriter(pretty=True)
        read_files_and_jobs = {}
        if not self._file_needs_to_be_created(output_stats_path):
            return
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for (lib_name, read_alignment_bam_path,
                 unaligned_reads_path) in zip(
                    lib_names, result_bam_paths, unaligned_reads_paths):
                read_aligner_stats = ReadAlignerStats()
                read_files_and_jobs[lib_name] = executor.submit(
                    read_aligner_stats.count, read_alignment_bam_path,
                    unaligned_reads_path)
        # Evaluate thread outcome
        self._check_job_completeness(read_files_and_jobs.values())
        read_files_and_stats = dict(
            [(lib_name, job.result())
             for lib_name, job in read_files_and_jobs.items()])
        raw_stat_data_writer.write(read_files_and_stats, output_stats_path)

    def _write_alignment_stat_table(self):
        """Manage the creation of the mapping statistic output table."""
        raw_stat_data_reader = RawStatDataReader()
        read_processing_stats = raw_stat_data_reader.read(
            self._paths.read_processing_stats_path)
        final_alignment_stats = raw_stat_data_reader.read(
            self._paths.read_alignments_stats_path)
        realignment_stats = None
        primary_aligner_stats = None
        if self._args.realign:
            primary_aligner_stats = raw_stat_data_reader.read(
                self._paths.primary_read_aligner_stats_path)
            realignment_stats = raw_stat_data_reader.read(
                self._paths.read_realigner_stats_path)
        read_aligner_stats_table = ReadAlignerStatsTable(
            read_processing_stats, final_alignment_stats,
            primary_aligner_stats,
            realignment_stats, self._lib_names,
            self._paths.read_alignment_stats_table_path, self._args.paired_end)
        read_aligner_stats_table.write()

    def _ref_ids_to_file(self, ref_seq_paths):
        """Translate the reference ID to file paths."""
        ref_ids_to_file = {}
        fasta_parser = FastaParser()
        for ref_seq_path in ref_seq_paths:
            ref_seq_file = os.path.basename(ref_seq_path)
            with open(ref_seq_path) as ref_seq_fh:
                ref_seq_id = fasta_parser.header_id(
                    fasta_parser.single_entry_file_header(ref_seq_fh))
                ref_ids_to_file[ref_seq_id] = ref_seq_file
        return ref_ids_to_file

    def create_coverage_files(self):
        """Create coverage files based on the read alignments.

        The coverages are calculated per replicon and the results are
        written to the output file. This might be slower but if all
        coverages are detmined at once the data structure will become
        too large when working with large reference sequences.

        """
        self._test_folder_existance(self._paths.required_coverage_folders())
        raw_stat_data_reader = RawStatDataReader()
        alignment_stats = [raw_stat_data_reader.read(
            self._paths.read_alignments_stats_path)]
        lib_names = list(alignment_stats[0].keys())
        was_paired_end_alignment = self._was_paired_end_alignment(lib_names)
        if not was_paired_end_alignment:
            self._paths.set_read_files_dep_file_lists_single_end(
                self._paths.get_read_files(), lib_names)
        else:
            self._paths.set_read_files_dep_file_lists_paired_end(
                self._paths.get_read_files(), lib_names)
        # Get number of aligned or number of uniquely aligned reads
        if not self._args.normalize_by_uniquely:
            aligned_counting = "no_of_aligned_reads"
        else:
            aligned_counting = "no_of_uniquely_aligned_reads"
        read_files_aligned_read_freq = dict([
            (read_file,
             round(attributes["stats_total"][aligned_counting]))
            for read_file, attributes in alignment_stats[0].items()])
        min_no_of_aligned_reads = float(min(
            read_files_aligned_read_freq.values()))
        # Run the generation of coverage in parallel
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for lib_name, bam_path in zip(
                    lib_names, self._paths.read_alignment_bam_paths):
                no_of_aligned_reads = float(
                    read_files_aligned_read_freq[lib_name])
                jobs.append(executor.submit(
                    self._create_coverage_files_for_lib,
                    lib_name, bam_path, no_of_aligned_reads,
                    min_no_of_aligned_reads))
        # Evaluate thread outcome
        self._check_job_completeness(jobs)

    def _all_coverage_file_exist(
        self, lib_name, strands, no_of_aligned_reads,
            min_no_of_aligned_reads):
        """Test the existance of all coverage file of a library"""
        files = []
        for strand in strands:
            files.append(self._paths.wiggle_file_raw_path(lib_name, strand))
            files.append(self._paths.wiggle_file_tnoar_norm_min_path(
                lib_name, strand, multi=min_no_of_aligned_reads,
                div=no_of_aligned_reads))
            files.append(self._paths.wiggle_file_tnoar_norm_mil_path(
                lib_name, strand, multi=1000000,
                div=no_of_aligned_reads))
        if not any([self._file_needs_to_be_created(file, quiet=True)
                    for file in files]):
            sys.stderr.write(
                "The files %s exists. Skipping their generation.\n" %
                ", " .join(files))
            return True
        return False

    def _create_coverage_files_for_lib(
        self, lib_name, bam_path, no_of_aligned_reads,
            min_no_of_aligned_reads):
        """Perform the coverage calculation for a given library."""
        if not self._args.non_strand_specific:
            strands = ["forward", "reverse"]
        else:
            strands = ["forward_and_reverse"]
        if self._all_coverage_file_exist(
                lib_name, strands, no_of_aligned_reads,
                min_no_of_aligned_reads):
            return
        read_count_splitting = True
        if self._args.skip_read_count_splitting:
            read_count_splitting = False
        coverage_calculator = CoverageCalculator(
            read_count_splitting=read_count_splitting,
            uniquely_aligned_only=self._args.unique_only,
            coverage_style=self._args.coverage_style,
            clip_length=self._args.clip_length,
            non_strand_specific=self._args.non_strand_specific)
        (coverage_writers_raw, coverage_writers_tnoar_min_norm,
         coverage_writers_tnoar_mil_norm) = self._wiggle_writers(
            lib_name, strands, no_of_aligned_reads, min_no_of_aligned_reads)
        for ref_seq, coverages in coverage_calculator.ref_seq_and_coverages(
                bam_path):
            for strand in strands:
                coverage_writers_raw[strand].write_replicons_coverages(
                    ref_seq, coverages[strand])
                coverage_writers_tnoar_min_norm[
                    strand].write_replicons_coverages(
                    ref_seq, coverages[strand],
                    factor=min_no_of_aligned_reads/no_of_aligned_reads)
                coverage_writers_tnoar_mil_norm[
                    strand].write_replicons_coverages(
                    ref_seq, coverages[strand],
                    factor=1000000/no_of_aligned_reads)
        for strand in strands:
            coverage_writers_raw[strand].close_file()

    def _wiggle_writers(self, lib_name, strands, no_of_aligned_reads,
                        min_no_of_aligned_reads):
        """Write the calculated coverages to wiggle files."""
        coverage_writers_raw = dict([(
            strand, WiggleWriter(
                "%s_%s" % (lib_name, strand),
                open(self._paths.wiggle_file_raw_path(lib_name, strand),
                     "w"))) for strand in strands])
        coverage_writers_tnoar_min_norm = dict([(
            strand, WiggleWriter(
                "%s_%s" % (lib_name, strand),
                open(self._paths.wiggle_file_tnoar_norm_min_path(
                    lib_name, strand, multi=min_no_of_aligned_reads,
                    div=no_of_aligned_reads), "w"))) for strand in strands])
        coverage_writers_tnoar_mil_norm = dict([(
            strand, WiggleWriter(
                "%s_%s" % (lib_name, strand),
                open(self._paths.wiggle_file_tnoar_norm_mil_path(
                    lib_name, strand, multi=1000000,
                    div=no_of_aligned_reads), "w"))) for strand in strands])
        return (coverage_writers_raw, coverage_writers_tnoar_min_norm,
                coverage_writers_tnoar_mil_norm)

    def _check_job_completeness(self, jobs):
        """Check the completness of each job in a list"""
        for job in concurrent.futures.as_completed(jobs):
            if job.exception():
                raise(job.exception())

    def quantify_gene_wise(self):
        """Manage the counting of aligned reads per gene."""
        self._test_folder_existance(
            self._paths.required_gene_quanti_folders())
        norm_by_alignment_freq = True
        norm_by_overlap_freq = True
        if self._args.no_count_split_by_alignment_no:
            norm_by_alignment_freq = False
        if self._args.no_count_splitting_by_gene_no:
            norm_by_overlap_freq = False
        raw_stat_data_reader = RawStatDataReader()
        alignment_stats = [raw_stat_data_reader.read(
            self._paths.read_alignments_stats_path)]
        lib_names = sorted(list(alignment_stats[0].keys()))
        annotation_files = self._paths.get_annotation_files()
        self._paths.set_annotation_paths(annotation_files)
        was_paired_end_alignment = self._was_paired_end_alignment(lib_names)
        if not was_paired_end_alignment:
            self._paths.set_read_files_dep_file_lists_single_end(
                self._paths.get_read_files(), lib_names)
        else:
            self._paths.set_read_files_dep_file_lists_paired_end(
                self._paths.get_read_files(), lib_names)
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            for lib_name, read_alignment_path in zip(
                    lib_names, self._paths.read_alignment_bam_paths):
                jobs.append(executor.submit(
                    self._quantify_gene_wise, lib_name,
                    read_alignment_path, norm_by_alignment_freq,
                    norm_by_overlap_freq, annotation_files))
        # Evaluate thread outcome
        self._check_job_completeness(jobs)
        self._gene_quanti_create_overview(
            annotation_files, self._paths.annotation_paths, lib_names)

    def _was_paired_end_alignment(self, lib_names):
        """Check if the mapping was done in paired- or single-end mode"""
        if len(lib_names) * 2 == len(self._paths.get_read_files()):
            return True
        return False

    def _quantify_gene_wise(
            self, lib_name, read_alignment_path,
            norm_by_alignment_freq,  norm_by_overlap_freq,
            annotation_files):
        """Perform the gene wise quantification for a given library."""
        gene_quanti_paths = [
            self._paths.gene_quanti_path(lib_name, annotation_file)
            for annotation_file in annotation_files]
        # Check if all output files for this library exist - if so
        # skip their creation
        if not any([self._file_needs_to_be_created(
                gene_quanti_path, quiet=True)
                for gene_quanti_path in gene_quanti_paths]):
            sys.stderr.write(
                "The file(s) %s exist(s). Skipping their/its generation.\n" %
                ", " .join(gene_quanti_paths))
            return
        gene_wise_quantification = GeneWiseQuantification(
            min_overlap=self._args.min_overlap,
            read_region=self._args.read_region,
            clip_length=self._args.clip_length,
            norm_by_alignment_freq=norm_by_alignment_freq,
            norm_by_overlap_freq=norm_by_overlap_freq,
            allowed_features_str=self._args.allowed_features,
            skip_antisense=self._args.skip_antisense,
            unique_only=self._args.unique_only)
        if norm_by_overlap_freq:
            gene_wise_quantification.calc_overlaps_per_alignment(
                read_alignment_path, self._paths.annotation_paths)
        for annotation_file, annotation_path in zip(
                annotation_files, self._paths.annotation_paths):
            gene_wise_quantification.quantify(
                read_alignment_path, annotation_path,
                self._paths.gene_quanti_path(
                    lib_name, annotation_file), self._args.pseudocounts)

    def _gene_quanti_create_overview(
            self, annotation_files, annotation_paths, lib_names):
        """Create an overview table of all gene quantification for all libs."""
        strand_specific = True
        if self._args.non_strand_specific:
                    strand_specific = False
        gene_wise_overview = GeneWiseOverview(
            allowed_features_str=self._args.allowed_features,
            skip_antisense=self._args.skip_antisense,
            strand_specific=strand_specific)
        path_and_name_combos = {}
        for annotation_file, annotation_path in zip(
                annotation_files, annotation_paths):
            path_and_name_combos[annotation_path] = []
            for read_file in lib_names:
                path_and_name_combos[annotation_path].append(
                    [read_file, self._paths.gene_quanti_path(
                        read_file, annotation_file)])
        if self._file_needs_to_be_created(
                self._paths.gene_wise_quanti_combined_path):
            gene_wise_overview.create_overview_raw_countings(
                path_and_name_combos, lib_names,
                self._paths.gene_wise_quanti_combined_path)
        if self._file_needs_to_be_created(
                self._paths.gene_wise_quanti_combined_rpkm_path):
            gene_wise_overview.create_overview_rpkm(
                path_and_name_combos, lib_names,
                self._paths.gene_wise_quanti_combined_rpkm_path,
                self._libs_and_total_num_of_aligned_reads())
        if self._file_needs_to_be_created(
                self._paths.gene_wise_quanti_combined_tnoar_path):
            gene_wise_overview.create_overview_norm_by_tnoar(
                path_and_name_combos, lib_names,
                self._paths.gene_wise_quanti_combined_tnoar_path,
                self._libs_and_total_num_of_aligned_reads())

    def _libs_and_total_num_of_aligned_reads(self):
        """Read the total number of reads per library."""
        with open(self._paths
                  .read_alignments_stats_path) as read_aligner_stats_fh:
            read_aligner_stats = json.loads(read_aligner_stats_fh.read())
        return dict([(lib, values["stats_total"]["no_of_aligned_reads"])
                     for lib, values in read_aligner_stats.items()])

    def _libs_and_total_num_of_uniquely_aligned_reads(self):
        """Read the total number of reads per library."""
        with open(self._paths
                  .read_alignments_stats_path) as read_aligner_stats_fh:
            read_aligner_stats = json.loads(read_aligner_stats_fh.read())
        return dict([(lib, values[
            "stats_total"]["no_of_uniquely_aligned_reads"])
            for lib, values in read_aligner_stats.items()])

    def compare_with_deseq(self):
        """Manage the pairwise expression comparison with DESeq."""
        self._test_folder_existance(
            self._paths.required_deseq_folders())
        arg_libs = [self._paths._clean_file_name(lib) for lib in
                    self._args.libs.split(",")]
        conditions = self._args.conditions.split(",")
        self._check_deseq_args(arg_libs, conditions)
        deseq_runner = DESeqRunner(
            arg_libs, conditions, self._paths.deseq_raw_folder,
            self._paths.deseq_extended_folder, self._paths.deseq_script_path,
            self._paths.deseq_pca_heatmap_path,
            self._paths.gene_wise_quanti_combined_path,
            self._paths.deseq_tmp_session_info_script,
            self._paths.deseq_session_info,
            self._args.fc_shrinkage_off,
            self._args.cooks_cutoff_off)
        deseq_runner.create_deseq_script_file()
        deseq_runner.write_session_info_file()
        deseq_runner.run_deseq()
        deseq_runner.merge_counting_files_with_results()

    def _check_deseq_args(self, arg_libs, conditions):
        """Test if the given arguments are sufficient."""
        if len(arg_libs) != len(conditions):
            self._write_err_msg_and_quit(
                "Error - The read library file list and condition list must "
                "have the same number of elements. You entered \n%s "
                "(= %s elements)\nand \n%s (= %s elements).\n" % (
                    self._args.libs, len(arg_libs), self._args.conditions,
                    len(conditions)))
        raw_stat_data_reader = RawStatDataReader()
        alignment_stats = [raw_stat_data_reader.read(
            self._paths.read_alignments_stats_path)]
        lib_names = list(alignment_stats[0].keys())
        if len(lib_names) != len(arg_libs):
            self._write_err_msg_and_quit(
                "The number of read libraries is lower or higher than "
                "expected. The following read libs are available: %s\nThe "
                "following read list string is suggested: \"%s\"\n" % (
                    ", ".join(lib_names), ",".join(lib_names)))
        for lib in lib_names:
            if lib not in arg_libs:
                self._write_err_msg_and_quit(
                    "The library \"%s\" is not present in your list of "
                    "libraries. Please add it.\n" % (lib))

    def _write_err_msg_and_quit(self, msg):
        """Write error message and close the program gracefully."""
        sys.stderr.write(msg)
        sys.exit(1)

    def viz_align(self):
        """Generate plots based on the read processing and mapping"""
        from reademptionlib.vizalign import AlignViz
        align_viz = AlignViz(
            self._paths.get_lib_names_single_end() if not self._args.paired_end
            else self._paths.get_lib_names_paired_end(),
            self._paths.read_processing_stats_path,
            self._paths.read_alignments_stats_path)
        align_viz.read_stat_files()
        align_viz.plot_input_read_length(
            self._paths.viz_align_input_read_length_plot_path)
        align_viz.plot_processed_read_length(
            self._paths.viz_align_processed_reads_length_plot_path)

    def viz_gene_quanti(self):
        """Generate plots based on the gene-wise read countings"""
        from reademptionlib.vizgenequanti import GeneQuantiViz
        gene_quanti_viz = GeneQuantiViz(
            self._paths.gene_wise_quanti_combined_path,
            self._paths.get_lib_names_single_end() if not self._args.paired_end
            else self._paths.get_lib_names_paired_end())
        gene_quanti_viz.parse_input_table()
        gene_quanti_viz.plot_correlations(
            self._paths.viz_gene_quanti_scatter_plot_path)
        gene_quanti_viz.plot_annotation_class_quantification(
            self._paths.viz_gene_quanti_rna_classes_plot_path)

    def viz_deseq(self):
        """Generate plots based on the DESeq analysis"""
        from reademptionlib.vizdeseq import DESeqViz
        deseq_path_template = (
            self._paths.deseq_raw_folder + "/deseq_comp_%s_vs_%s.csv")
        deseq_viz = DESeqViz(
            self._paths.deseq_script_path, deseq_path_template,
            max_pvalue=self._args.max_pvalue)
        deseq_viz.create_scatter_plots(
            self._paths.viz_deseq_scatter_plot_path)
        deseq_viz.create_volcano_plots(
            self._paths.viz_deseq_volcano_plot_path,
            self._paths.viz_deseq_volcano_plot_adj_path)



import pysam
import sys


class CrossAlignFilter(object):
    
    def __init__(self, input_bam, output_bam, output_crossmapped_reads,
                 orgs_and_replicon_ids):
        self._input_bam = input_bam
        self._output_bam = output_bam
        self._output_crossmapped_reads = output_crossmapped_reads
        self._orgs_and_replicon_ids = orgs_and_replicon_ids
        self._crossmapped_reads = set()
        self.no_of_crossmapped_reads = None
        
    def determine_crossmapped_reads(self):
        """Find reads that are mapped to sequences of different species.

        The comparison is performed replicon wise in order to reduce
        the memory footprint.
        """
        self._check_replicon_existance()
        done_replicon_comparison = []
        with pysam.Samfile(self._input_bam) as bam:
            for org, replicon_ids in self._orgs_and_replicon_ids.items():
                for replicon_id in replicon_ids:
                    self._read_ids = set()
                    # First, collect the ids of the aligned reads of
                    # this replicon
                    for alignment in bam.fetch(reference=replicon_id):
                        self._read_ids.add(alignment.qname)
                    # Then compare them to the alignments of each
                    # replicon of the other organism(s)
                    for comp_org, comp_replicon_ids in (
                            self._orgs_and_replicon_ids.items()):
                        # Only compare replicons of different species
                        if org == comp_org:
                            continue
                        for comp_replicon_id in comp_replicon_ids:
                            comparison = sorted([replicon_id,
                                                 comp_replicon_id])
                            # Check if comparison of the two replicons
                            # has been done already
                            if comparison in done_replicon_comparison:
                                continue
                            done_replicon_comparison.append(comparison)
                            # Compare all read ids of the comparison
                            # replicon to the query replicon read ids
                            for alignment in bam.fetch(
                                    reference=comp_replicon_id):
                                if alignment.qname in self._read_ids:
                                    self._crossmapped_reads.add(
                                        alignment.qname)
        self.no_of_crossmapped_reads = len(self._crossmapped_reads)
        # Write list of cross mapped read to file
        with open(self._output_crossmapped_reads, "w") as output_list_fh:
            output_list_fh.write("\n".join(self._crossmapped_reads) + "\n")

    def _check_replicon_existance(self):
        found_all = True
        with pysam.Samfile(self._input_bam) as bam:
            for replicon_ids in self._orgs_and_replicon_ids.values():
                for replicon_id in replicon_ids:
                    if not replicon_id in bam.references:
                        sys.stderr.write(
                            "\"%s\" not found in BAM header.\n" % replicon_id)
                        found_all = False
        if not found_all:
            raise RepliconIdNotInBam
        
    def write_crossmapping_free_bam(self):
        with pysam.Samfile(self._input_bam) as input_bam:
            with pysam.Samfile(self._output_bam, "wb",
                               header=input_bam.header) as output_bam:
                for alignment in input_bam.fetch():
                    if not alignment.qname in self._crossmapped_reads:
                        output_bam.write(alignment)
        pysam.index(self._output_bam)


class RepliconIdNotInBam(BaseException):
    pass




from reademptionlib.lack import Lack


class ReadRealigner(object):
    """An abstraction layer for different short read realigners."""

    def __init__(self, lack_bin, show_progress):
        self.lack = Lack(lack_bin, show_progress=show_progress)
    
    def run_alignment(self, query_fasta_path, query_sam_path, ref_seq_paths,
                      output_sam_path, nomatch_path, threads, accuracy):
        self.lack.realign_reads(query_fasta_path,
                                query_sam_path,
                                ref_seq_paths,
                                output_sam_path,
                                nonmatch_path=nomatch_path, threads=threads,
                                accuracy=accuracy)



import os
import concurrent.futures
from reademptionlib.paths import Paths
from reademptionlib.wiggle import WiggleWriter


class Controller_TK(object):

    def __init__(self, args):
        """Create an instance."""
        self._args = args
        self._paths = Paths

    def viz_align_TK(self):
        """Generate plots based on the read processing and mapping"""
        from reademptionlib.vizalign2 import AlignViz2
        align_viz_tk = AlignViz2()
        if self._args.input_align:
            align_viz_tk.alignment_stats(
                str(self._args.input_align),
                str(self._args.output_align),
                str(self._args.output_align))
        if self._args.input_process:
            align_viz_tk.process_stats(
                str(self._args.input_process),
                str(self._args.output_process),
                str(self._args.output_process))
        
    def get_and_submit_bam_stats(self):
        from reademptionlib.readalignerstats import ReadAlignerStats
        get_bam_stats = ReadAlignerStats()
        total_aligned_reads = []
        total_uniquely_aligned_reads = []
        all_aligned_reads = []
        for sample_or_stats_total, sub_dic in (
                get_bam_stats.count(self._args.input_coverage, "NA")).items():
            for stats_total, value in sub_dic.items():
                if sample_or_stats_total == 'stats_total':
                    if stats_total == 'no_of_uniquely_aligned_reads':
                        total_uniquely_aligned_reads.append(value)
                    if stats_total == 'no_of_aligned_reads':
                        total_aligned_reads.append(value)
                else:
                    for align_type, Nr in value.items():
                        if align_type == 'no_of_aligned_reads':
                            all_aligned_reads.append(Nr)
        min_no_of_aligned_reads = min(all_aligned_reads)
        if not self._args.normalize_by_uniquely:
            no_of_aligned_reads = "".join([
                '{:.1f}'.format(x) for x in total_aligned_reads])
        else:
            no_of_aligned_reads = "".join([
                '{:.1f}'.format(x) for x in total_uniquely_aligned_reads])
        jobs = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=self._args.processes) as executor:
            jobs.append(executor.submit(
                self.create_coverage_files_TK, float(no_of_aligned_reads),
                float(min_no_of_aligned_reads)))
        self._check_job_completeness(jobs)

    def create_coverage_files_TK(
            self, no_of_aligned_reads, min_no_of_aligned_reads):
        if not self._args.non_strand_specific:
            strands = ["forward", "reverse"]
        else:
            strands = ["forward_and_reverse"]
        if self._args.skip_read_count_splitting:
            read_count_splitting = False
        else:
            read_count_splitting = True
        from reademptionlib.coveragecalculator import CoverageCalculator
        coverage_calculator = CoverageCalculator(
            read_count_splitting=read_count_splitting,
            uniquely_aligned_only=self._args.unique_only,
            coverage_style=self._args.coverage_style,
            clip_length=self._args.clip_length,
            non_strand_specific=self._args.non_strand_specific)
        (coverage_writers_raw, coverage_writers_tnoar_min_norm,
         coverage_writers_tnoar_mil_norm) = self._wiggle_writers(
             strands, no_of_aligned_reads, min_no_of_aligned_reads)
        for ref_seq, coverages in coverage_calculator.ref_seq_and_coverages(
                self._args.input_coverage):
            for strand in strands:
                coverage_writers_raw[strand].write_replicons_coverages(
                    ref_seq, coverages[strand])
            for strand in strands:
                coverage_writers_tnoar_min_norm[
                    strand].write_replicons_coverages(
                    ref_seq, coverages[strand],
                    factor=min_no_of_aligned_reads/no_of_aligned_reads)
            for strand in strands:
                coverage_writers_tnoar_mil_norm[
                    strand].write_replicons_coverages(
                    ref_seq, coverages[strand],
                    factor=self._args.factor/no_of_aligned_reads)
        for strand in strands:
            coverage_writers_raw[strand].close_file()

    def _wiggle_writers(self, strands, no_of_aligned_reads,
                        min_no_of_aligned_reads):
        if self._args.output_coverage is None:
            filename = self.get_filename(self._args.input_coverage)
            if os.path.dirname(self._args.input_coverage) == "":
                output_path = "."
            else:
                output_path = os.path.dirname(self._args.input_coverage)
        else:
            filename = self.get_filename(self._args.output_coverage)
            if os.path.dirname(self._args.output_coverage) == "":
                output_path = "."
            else:
                output_path = os.path.dirname(self._args.output_coverage)
        if self._args.trackname is None:
            trackname = filename
        else:
            trackname = self.get_filename(self._args.trackname)
        coverage_writers_raw = dict([(
            strand, WiggleWriter(
                "%s_%s" % (trackname, strand),
                open("%s/%s_%s-raw.wig" % (output_path, filename, strand),
                     "w"))) for strand in strands])
        coverage_writers_tnoar_min_norm = dict([(
            strand, WiggleWriter(
                "%s_%s" % (trackname, strand),
                open(
                    "%s/%s_%s-tnoar_mil_norm_multi_by_%.1f_div_by_%.1f.wig" % (
                        output_path, filename, strand,
                        min_no_of_aligned_reads, no_of_aligned_reads),
                    "w"))) for strand in strands])
        coverage_writers_tnoar_mil_norm = dict([(
            strand, WiggleWriter(
                "%s_%s" % (trackname, strand),
                open(
                    "%s/%s_%s-tnoar_mil_norm_multi_by_%.1f_div_by_%.1f.wig" % (
                        output_path, filename, strand, self._args.factor,
                        no_of_aligned_reads),
                    "w"))) for strand in strands])
        return(coverage_writers_raw, coverage_writers_tnoar_min_norm,
               coverage_writers_tnoar_mil_norm)

    def get_filename(self, path_file):
        return(os.path.splitext(os.path.basename(path_file))[0])

    def _check_job_completeness(self, jobs):
        """Check the completness of each job in a list"""
        for job in concurrent.futures.as_completed(jobs):
            if job.exception():
                raise(job.exception())
            





import csv
import sys


class Gff3Parser(object):
    """
    A format description can be found at:
    http://genome.ucsc.edu/FAQ/FAQformat.html#format3
    http://www.sequenceontology.org/gff3.shtml

    a validator can be found here:
    http://modencode.oicr.on.ca/cgi-bin/validate_gff3_online
    """

    def entries(self, input_gff_fh):
        """
        """
        for entry_dict in csv.DictReader(
            input_gff_fh, delimiter="\t",
            fieldnames=["seq_id", "source", "feature", "start",
                        "end", "score", "strand", "phase", "attributes"]):
            if entry_dict["seq_id"].startswith("#"):
                continue
            try:
                yield(self._dict_to_entry(entry_dict))
            except:
                sys.stderr.write(
                    "Error! Please make sure that you use GFF3 formated "
                    "annotation files. GTF2/GTF is not valid and the usage "
                    "of that format is not recommended anymore (see "
                    "http://www.sequenceontology.org/gff3.shtml for more "
                    "information).\n")
                sys.exit(0)
    
    def _dict_to_entry(self, entry_dict):
        return Gff3Entry(entry_dict)


class Gff3Entry(object):

    def __init__(self, entry_dict):
        self.seq_id = entry_dict["seq_id"]
        self.source = entry_dict["source"]
        self.feature = entry_dict["feature"]
        # 1-based coordinates
        # Make sure that start <= end
        start, end = sorted([int(entry_dict["start"]), int(entry_dict["end"])])
        self.start = start
        self.end = end
        self.score = entry_dict["score"]
        self.strand = entry_dict["strand"]
        self.phase = entry_dict["phase"]
        self.attributes = self._attributes(entry_dict["attributes"])
        self.attribute_string = entry_dict["attributes"]
    
    def _attributes(self, attributes_string):
        """Translate the attribute string to dictionary"""
        if attributes_string is None:
            return({})
        if attributes_string.endswith(";"):
            attributes_string = attributes_string[:-1]
        return dict(
            [key_value_pair.split("=")
             for key_value_pair in attributes_string.split(";")])

    def __str__(self):
        return "\t".join([str(field) for field in [
            self.seq_id, self.source, self.feature, self.start,
            self.end, self.score, self.strand, self.phase,
            self.attribute_string]])



from io import StringIO
import unittest
import sys
sys.path.append(".")
from reademptionlib.gff3 import Gff3Parser, Gff3Entry

class TestGff3Parser(unittest.TestCase):

    def setUp(self):
        self.gff3parser = Gff3Parser()
        self.example_data = ExampleData()

    def test_entries(self):
        gff_fh = StringIO(self.example_data.gff_content_1)
        entry_seq_ids = [entry.seq_id for entry 
                        in self.gff3parser.entries(gff_fh)]
        self.assertListEqual(
            ['foo01', 'foo01', 'foo02', 'foo02'], entry_seq_ids)
        
class TestGff3Entry(unittest.TestCase):

    def test_entry_creation_1(self):
        gff3entry = Gff3Entry({
                "seq_id" : "IDfoobar2342",
                "source" : "a_prog",
                "feature" : "gene",
                "start" : "1984",
                "end" : "2001",
                "score" : "1323",
                "strand" : "+",
                "phase" : ".",
                "attributes" : "ID=foo00001;Name=foo"})
        self.assertEqual(gff3entry.seq_id, "IDfoobar2342")
        self.assertEqual(gff3entry.source, "a_prog")
        self.assertEqual(gff3entry.feature, "gene")
        self.assertEqual(gff3entry.start, 1984)
        self.assertEqual(gff3entry.end, 2001)
        self.assertEqual(gff3entry.score, "1323")
        self.assertEqual(gff3entry.strand, "+")
        self.assertEqual(gff3entry.phase, ".")
        self.assertDictEqual(
            gff3entry.attributes, {"ID": "foo00001", "Name": "foo"})

    def test_entry_creation_1(self):
        gff3entry = Gff3Entry({
                "seq_id" : "IDfoobar2342",
                "source" : "a_prog",
                "feature" : "gene",
                "start" : "1984",
                "end" : "2001",
                "score" : "1323",
                "strand" : "+",
                "phase" : ".",
                "attributes" : "ID=foo00001;Name=foo"})
        self.assertEqual(gff3entry.seq_id, "IDfoobar2342")
        self.assertEqual(gff3entry.source, "a_prog")
        self.assertEqual(gff3entry.feature, "gene")
        self.assertEqual(gff3entry.start, 1984)
        self.assertEqual(gff3entry.end, 2001)
        self.assertEqual(gff3entry.score, "1323")
        self.assertEqual(gff3entry.strand, "+")
        self.assertEqual(gff3entry.phase, ".")
        self.assertDictEqual(
            gff3entry.attributes, {"ID": "foo00001", "Name": "foo"})
        self.assertEqual(gff3entry.attribute_string, "ID=foo00001;Name=foo")

    def test_entry_creation_2(self):
        gff3entry = Gff3Entry({
                "seq_id" : "accession_424",
                "source" : "make",
                "feature" : "CDS",
                "start" : "101",
                "end" : "199",
                "score" : "0.888",
                "strand" : "-",
                "phase" : "1",
                "attributes" : "ID=blub1;Name=wow01"})
        self.assertEqual(gff3entry.seq_id, "accession_424")
        self.assertEqual(gff3entry.source, "make")
        self.assertEqual(gff3entry.feature, "CDS")
        self.assertEqual(gff3entry.start, 101)
        self.assertEqual(gff3entry.end, 199)
        self.assertEqual(gff3entry.score, "0.888")
        self.assertEqual(gff3entry.strand, "-")
        self.assertEqual(gff3entry.phase, "1")
        self.assertDictEqual(
            gff3entry.attributes, {"ID": "blub1", "Name": "wow01"})
        self.assertEqual(gff3entry.attribute_string, "ID=blub1;Name=wow01")

    def test_entry_creation_3(self):
        """Test that start and end are sorted. """
        gff3entry = Gff3Entry({
                "seq_id" : "accession_111",
                "source" : "make",
                "feature" : "CDS",
                "start" : "1000",
                "end" : "1",
                "score" : "",
                "strand" : "-",
                "phase" : "",
                "attributes" : "locus_tag=boing"})
        self.assertEqual(gff3entry.start, 1)
        self.assertEqual(gff3entry.end, 1000)

    def test_string_creation(self):
        gff3entry = Gff3Entry({
                "seq_id" : "accession_111",
                "source" : "make",
                "feature" : "CDS",
                "start" : "1000",
                "end" : "1",
                "score" : "0.5",
                "strand" : "-",
                "phase" : ".",
                "attributes" : "locus_tag=boing;note=zoong"})
        self.assertEqual(
            str(gff3entry), 
            "accession_111\tmake\tCDS\t1\t1000\t0.5\t-\t.\t"
            "locus_tag=boing;note=zoong")

class ExampleData(object):
    
    gff_content_1 ="""##gff-version 3
foo01	ta_prog	gene	100	200	.	+	.	ID=bar01;Name=blub
foo01	a_prog	gene	400	500	.	+	.	ID=bar02;Name=limbo
foo02	a_prog	gene	400	500	.	+	.	ID=bar03
foo02	a_prog	CDS	400	500	.	+	.	ID=bar03;Name=bla
"""

if __name__ == "__main__":
	unittest.main()





class ParameterLog(object):

    def __init__(self, log_file):
        self.log_file = log_file
        self.delimiter = ": "

class ParameterLogger(ParameterLog):

    def add_paramter(self, parameter, value):
        log_fh = open(self.log_file, "a")
        log_fh.write(self.delimiter.join([parameter, value]) + "\n")
        log_fh.close()

class ParameterLogReader(ParameterLog):
        
    def read_parameters(self):
        parameters = {}
        for line in open(self.log_file):
            paramter, value = line[:-1].split(self.delimiter)
            parameters[parameters] = value



import sys
from collections import defaultdict
from functools import reduce
from reademptionlib.fasta import FastaParser
import pysam


class ReadAlignerStats(object):

    def __init__(self):
        self.fasta_parser = FastaParser()

    def count(self, read_alignment_result_bam_path, unaligned_reads_path):
        self._stats = {}
        self._count_aligned_reads_and_alignments(
            read_alignment_result_bam_path)
        self._count_unaligned_reads(unaligned_reads_path)
        return self._stats

    def _count_unaligned_reads(self, unaligned_read_paths):
        
        with open(unaligned_read_paths) as fasta_fh:
            self._stats["stats_total"][
                "no_of_unaligned_reads"] = self._count_fasta_entries(fasta_fh)

    def _count_fasta_entries(self, fasta_fh):
        return reduce(lambda x, y: x + 1,
                      self.fasta_parser.entries(fasta_fh), 0)

    def _count_aligned_reads_and_alignments(
            self, read_alignment_result_bam_path):
        bam = pysam.Samfile(read_alignment_result_bam_path)
        stats_per_ref = defaultdict(dict)
        no_of_hits_per_read_freq = {}
        for ref_id in bam.references:
            self._init_counting_dict(stats_per_ref, ref_id)
        for entry in bam.fetch():
            ref_id = bam.getrname(entry.tid)
            try:
                self._count_alignment(
                    entry, ref_id, stats_per_ref, no_of_hits_per_read_freq)
            except KeyError:
                sys.stderr.write(
                    "SAM entry with unspecified reference found! Stoping\n")
                sys.exit(2)
        self._stats["stats_per_reference"] = stats_per_ref
        for ref_id, stats in stats_per_ref.items():
            stats_per_ref[ref_id][
                "no_of_hits_per_read_and_freqs"] = self._calc_down_to_read(
                stats_per_ref[ref_id]["no_of_hits_per_read_and_freqs"])
        self._stats["stats_total"] = self._sum_countings(stats_per_ref)

    def _sum_countings(self, stats_per_ref):
        total_stats = {}
        for ref_id, stats in stats_per_ref.items():
            for attribute, value in stats.items():
                if type(value) is int or type(value) is float:
                    total_stats.setdefault(attribute, 0)
                    total_stats[attribute] += value
                elif type(value) is dict:
                    total_stats.setdefault(attribute, {})
                    for value_int, freq in value.items():
                        total_stats[attribute].setdefault(value_int, 0)
                        total_stats[attribute][value_int] += freq
        return total_stats

    def _calc_down_to_read(self, no_of_hits_per_read_freq):
        """As the frequencies were determined via the alignments we need
        to normalized each frequency value down to the read by
        dividing the frequencig by the number of hits per read.
        """
        return dict((no_of_hits_per_read, freq/no_of_hits_per_read)
                    for no_of_hits_per_read, freq in
                    no_of_hits_per_read_freq.items())

    def _init_counting_dict(self, stats_per_ref, ref_id):
        stats_per_ref[ref_id] = defaultdict(float)
        stats_per_ref[ref_id]["no_of_alignments"]
        stats_per_ref[ref_id]["no_of_aligned_reads"]
        stats_per_ref[ref_id]["no_of_split_alignments"]
        stats_per_ref[ref_id]["no_of_uniquely_aligned_reads"]
        stats_per_ref[ref_id][
            "alignment_length_and_freqs"] = defaultdict(int)
        stats_per_ref[ref_id][
            "no_of_hits_per_read_and_freqs"] = defaultdict(int)

    def _count_alignment(self, entry, ref_id, stats_per_ref,
                         no_of_hits_per_read_freq):
        entry_tags_dict = dict(entry.tags)
        no_of_hits = entry_tags_dict["NH"]
        # Consider split reads
        no_of_splits = float(entry_tags_dict.get("XL", 1))
        stats_per_ref[ref_id]["no_of_hits_per_read_and_freqs"][
            no_of_hits] += 1
        if "XL" in entry_tags_dict:
            stats_per_ref[ref_id]["no_of_split_alignments"] += 1.0/no_of_splits
        stats_per_ref[ref_id]["no_of_alignments"] += 1.0/no_of_splits
        stats_per_ref[
            ref_id]["no_of_aligned_reads"] += 1.0/(
            float(no_of_hits) * no_of_splits)
        if no_of_hits == 1:
            stats_per_ref[ref_id][
                "no_of_uniquely_aligned_reads"] += 1.0/no_of_splits
        stats_per_ref[ref_id][
            "alignment_length_and_freqs"][entry.alen] += 1



import sys
import os
import unittest
sys.path.append(".")
from reademptionlib.coveragecalculator import CoverageCalculator
from io import StringIO
import pysam

class TestCoverageCalculator(unittest.TestCase):

    def setUp(self):
        self.coverage_calculator = CoverageCalculator()
        self.example_data = ExampleData()
        self._sam_bam_prefix = "dummy"

    def tearDown(self):
        for suffix in [".sam", ".bam", ".bam.bai"]:
            if os.path.exists(self._sam_bam_prefix + suffix) is True:
                os.remove(self._sam_bam_prefix + suffix)

    def _generate_bam_file(self, sam_content, file_prefix):
        sam_file = "{}.sam".format(file_prefix)
        bam_file = "{}.bam".format(file_prefix)
        sam_fh = open(sam_file, "w")
        sam_fh.write(sam_content)
        sam_fh.close()
        pysam.view(
            "-Sb", "-o{}".format(bam_file), sam_file, catch_stdout=False)
        pysam.index(bam_file)
        self._bam = pysam.Samfile(bam_file)

    def test_init_coverage_list(self):
        self.coverage_calculator._init_coverage_list(10)
        self.assertListEqual(
            sorted(self.coverage_calculator._coverages.keys()),
            ["forward", "reverse"])
        self.assertListEqual(
            self.coverage_calculator._coverages["forward"].tolist(), [0.0] * 10)
        self.assertListEqual(
            self.coverage_calculator._coverages["reverse"].tolist(), [0.0] * 10)

    def test_calc_coverage_1(self):
        """Check correct start at first list element"""
        self._generate_bam_file(
            self.example_data.sam_content_1, self._sam_bam_prefix)
        self.coverage_calculator._init_coverage_list(self._bam.lengths[0])
        self.coverage_calculator._calc_coverage("chrom", self._bam)
        self.assertEqual(
            len(self.coverage_calculator._coverages["forward"]), 1500)
        self.assertListEqual(
            self.coverage_calculator._coverages["forward"][0:15].tolist(),
            [5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0, 5.0,
             0.0, 0.0, 0.0, 0.0, 0.0])
        self.assertListEqual(
            self.coverage_calculator._coverages["reverse"][0:15].tolist(),
            [-5.0, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0, -5.0,
             0.0, 0.0, 0.0, 0.0, 0.0])

    def test_calc_coverage_2(self):
        """Consider how often a read is mapped. Mappings of reads that
        are aligned to several location contribute only fractions to
        the counting.

        """
        self._generate_bam_file(
            self.example_data.sam_content_2, self._sam_bam_prefix)
        self.coverage_calculator._init_coverage_list(self._bam.lengths[0])
        self.coverage_calculator._calc_coverage("chrom", self._bam)
        self.assertListEqual(
            self.coverage_calculator._coverages[
                "forward"][0:15].tolist(),
            [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5,
             0.0, 0.0, 0.0, 0.0, 0.0])

    def test_calc_coverage_3(self):
        """If read_count_splitting is set to False then every
        mapping is counted as one to each of the matching position
        independent how often its read is mapped in in total.
        """
        self.coverage_calculator = CoverageCalculator(
            read_count_splitting=False)
        self._generate_bam_file(
            self.example_data.sam_content_2, self._sam_bam_prefix)
        self.coverage_calculator._init_coverage_list(self._bam.lengths[0])
        self.coverage_calculator._calc_coverage("chrom", self._bam)
        self.assertListEqual(
            self.coverage_calculator._coverages[
                "forward"][0:15].tolist(),
            [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
             0.0, 0.0, 0.0, 0.0, 0.0])

    def test_calc_coverage_4(self):
        """If uniqueley_aligned_only is True skip any mapping of read
        that are aligned to more than on location.
        """
        self.coverage_calculator = CoverageCalculator(
            uniquely_aligned_only=True)
        self._generate_bam_file(
            self.example_data.sam_content_3, self._sam_bam_prefix)
        self.coverage_calculator._init_coverage_list(self._bam.lengths[0])
        self.coverage_calculator._calc_coverage("chrom", self._bam)
        self.assertListEqual(
            self.coverage_calculator._coverages[
                "forward"][0:15].tolist(),
            [3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0,
             0.0, 0.0, 0.0, 0.0, 0.0])

    def test_calc_coverage_5(self):
        """If first_base_only is True only the first nucleotide of a
        mapping is considered.
        """
        self.coverage_calculator = CoverageCalculator(
            coverage_style="first_base_only")
        self._generate_bam_file(
            self.example_data.sam_content_1, self._sam_bam_prefix)
        self.coverage_calculator._init_coverage_list(self._bam.lengths[0])
        self.coverage_calculator._calc_coverage("chrom", self._bam)
        self.assertListEqual(
            self.coverage_calculator._coverages[
                "forward"][0:15].tolist(),
                [5.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                 0.0, 0.0, 0.0, 0.0, 0.0])
        self.assertListEqual(
            self.coverage_calculator._coverages[
                "reverse"][0:15].tolist(),
            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -5.0,
             0.0, 0.0, 0.0, 0.0, 0.0])

class ExampleData(object):

    sam_content_1 = """@HD	VN:1.0
@SQ	SN:chrom	LN:1500
@SQ	SN:plasmid1	LN:100
@SQ	SN:plasmid2	LN:200
myread:001	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:002	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:003	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:004	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:005	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:006	16	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:007	16	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:008	16	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:009	16	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:010	16	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
"""
    sam_content_2 = """@HD	VN:1.0
@SQ	SN:chrom	LN:1500
@SQ	SN:plasmid1	LN:100
@SQ	SN:plasmid2	LN:200
myread:001	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:2	XI:i:1	XA:Z:Q
"""

    sam_content_3 = """@HD	VN:1.0
@SQ	SN:chrom	LN:1500
@SQ	SN:plasmid1	LN:100
@SQ	SN:plasmid2	LN:200
myread:001	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:002	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:003	0	chrom	1	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:1	XI:i:1	XA:Z:Q
myread:004	0	chrom	5	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:9	XI:i:1	XA:Z:Q
myread:005	0	chrom	5	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:9	XI:i:1	XA:Z:Q
myread:006	0	chrom	5	255	10M	*	0	0	GTGGACAACC	*	NM:i:1	MD:Z:11T3	NH:i:9	XI:i:1	XA:Z:Q
"""

if __name__ == "__main__":
    unittest.main()



import json


class RawStatDataWriter(object):

    def __init__(self, pretty=False):
        self._pretty = pretty

    def write(self, input_object, output_path):
        with open(output_path, "w") as output_fh:
            self._write_json(input_object, output_fh)

    def _write_json(self, input_object, output_fh):
        if self._pretty is True:
            indent = 4
        else:
            indent = None
        output_fh.write(json.dumps(input_object, indent=indent))


class RawStatDataReader(object):
            
    def read(self, input_file):
        with open(input_file) as input_fh:
            data = self._read(input_fh)
            input_fh.close()
        return data

    def _read(self, input_fh):
        return json.loads(input_fh.read())



import os
import unittest
import shutil
import sys
sys.path.append(".")
from reademptionlib.paths import Paths

class TestPaths(unittest.TestCase):

    test_folder = "/tmp/test"
    test_files = ["foo.fa", "bar.fa"]
    test_lib_names = ["foo", "bar"]

    def setUp(self):
        self.paths = Paths(base_path=self.test_folder)
        self.folder_names = [
            self.paths.input_folder,
            self.paths.output_folder,
            self.paths.align_report_folder,
            self.paths.raw_stat_data_folder,
            self.paths.read_fasta_folder,
            self.paths.ref_seq_folder,
            self.paths.annotation_folder,
            self.paths.read_alignment_index_folder,
            self.paths.read_alignments_folder,
            self.paths.processed_reads_folder,
            self.paths.unaligned_reads_folder,
            self.paths.coverage_raw_folder,
            self.paths.coverage_tnoar_min_norm_folder,
            self.paths.coverage_tnoar_mil_norm_folder,
            self.paths.gene_quanti_base_folder,
            self.paths.gene_wise_quanti_combined_path]
        self.static_files = [
            self.paths.read_processing_stats_path,
            self.paths.read_alignments_stats_path,
            self.paths.read_file_stats,
            self.paths.read_alignment_stats_table_path,
            self.paths.ref_seq_file_stats,
            self.paths.index_path]

    def tearDown(self):
        self._remove_folder_if_exists(self.test_folder)

    def test_set_folder_names(self):
        self.paths._set_folder_names()
        for folder_name in self.folder_names:
            assert(folder_name != '')
            self.assertEqual(self.folder_names.count(folder_name), 1)

    def test_set_folder_names_with_base_path(self):
        self.paths._set_folder_names()
        for folder_name in self.folder_names:
            assert(folder_name != '')
            self.assertEqual(self.folder_names.count(folder_name), 1)

    def test_set_files(self):
        self.paths._set_static_files
        for file_name in self.static_files:
            assert(file_name != '')
            self.assertEqual(self.static_files.count(file_name), 1)

    def test_get_sorted_folder_content(self):
        self._create_folder_with_files(self.test_folder, self.test_lib_names)
        self.assertEqual(
            self.paths._get_sorted_folder_content(self.test_folder),
            sorted(self.test_lib_names))
        self._remove_folder_if_exists(self.test_folder)

    def test_required_folders(self):
        self.assertEqual(len(self.paths.required_folders()), 25)

    def test_get_read_files(self):
        self.paths.read_fasta_folder = self.test_folder
        self._create_folder_with_files(self.test_folder, self.test_lib_names)
        self.assertEqual(self.paths.get_read_files(), sorted(self.test_lib_names))
        self._remove_folder_if_exists(self.test_folder)

    def test_get_ref_seq_files(self):
        self.paths.ref_seq_folder = self.test_folder
        self._create_folder_with_files(self.test_folder, self.test_lib_names)
        self.assertEqual(self.paths.get_ref_seq_files(),
                         sorted(self.test_lib_names))
        self._remove_folder_if_exists(self.test_folder)

    def test_get_annotation_files(self):
        self.paths.annotation_folder = self.test_folder
        self._create_folder_with_files(self.test_folder, self.test_lib_names)
        self.assertEqual(self.paths.get_annotation_files(),
                         sorted(self.test_lib_names))
        self._remove_folder_if_exists(self.test_folder)

    def test_set_read_files_dep_file_lists(self):
        self.paths.set_read_files_dep_file_lists_single_end(
            self.test_files, self.test_lib_names)
        for path_list in  [
            self.paths.read_paths, self.paths.processed_read_paths,
            self.paths.primary_read_aligner_sam_paths,
            self.paths.unaligned_reads_paths]:
            assert(isinstance(path_list, list))

    def test_path_list_without_appendix(self):
        self.assertEqual(
            self.paths._path_list(self.test_folder, self.test_lib_names),
            ['/tmp/test/foo', '/tmp/test/bar'])

    def test_path_list_with_appendix(self):
        self.assertEqual(
            self.paths._path_list(
                self.test_folder, self.test_lib_names, appendix=".boing"),
            ['/tmp/test/foo.boing', '/tmp/test/bar.boing'])

    def test_set_ref_seq_paths(self):
        self.paths.set_ref_seq_paths(self.test_files)
        self.assertEqual(
            self.paths.ref_seq_paths,
            ["%s/%s" % (self.paths.ref_seq_folder, file) for
             file in self.test_files])

    def _create_folder_with_files(self, folder, file_names):
        os.mkdir(self.test_folder)
        for file_name in ["foo", "bar"]:
            open("%s/%s" % (self.test_folder, file_name), "w").close()

    def _remove_folder_if_exists(self, folder):
        if os.path.exists(folder):
            shutil.rmtree(folder)

if __name__ == "__main__":
    unittest.main()



import csv
import sys
import os
from subprocess import call


class DESeqRunner(object):

    def __init__(
            self, libs, conditions, deseq_raw_folder, deseq_extended_folder,
            deseq_script_path, deseq_pca_heatmap_path,
            gene_wise_quanti_combined_path,
            deseq_tmp_session_info_script, deseq_session_info,
            fc_shrinkage_off,
            cooks_cutoff_off=False):
        self._libs = libs
        self._conditions = conditions
        self._deseq_raw_folder = deseq_raw_folder
        self._deseq_extended_folder = deseq_extended_folder
        self._deseq_script_path = deseq_script_path
        self._deseq_pca_heatmap_path = deseq_pca_heatmap_path
        self._gene_wise_quanti_combined_path = gene_wise_quanti_combined_path
        self._deseq_tmp_session_info_script = deseq_tmp_session_info_script
        self._deseq_session_info = deseq_session_info
        self._cooks_cutoff_off = cooks_cutoff_off
        self._first_data_column = 11
        self._fc_shrinkage_off = fc_shrinkage_off

    def write_session_info_file(self):
        with open(self._deseq_tmp_session_info_script, "w") as tmp_r_script_fh:
            tmp_r_script_fh.write("library('DESeq2')\nsessionInfo()\n")
        with open(self._deseq_session_info, "w") as session_info_fh:
            with open(os.devnull, "w") as devnull:
                call(["Rscript", self._deseq_tmp_session_info_script],
                     stdout=session_info_fh, stderr=devnull)
        os.remove(self._deseq_tmp_session_info_script)

    def create_deseq_script_file(self):
        libs_to_conditions = dict([
            (lib, condition) for lib, condition in
            zip(self._libs, self._conditions)])
        head_row = open(
            self._gene_wise_quanti_combined_path).readline()[:-1].split("\t")
        libs = head_row[self._first_data_column-1:]
        libs_str = ",".join(["'%s'" % lib for lib in libs])
        conditions = [libs_to_conditions[lib] for lib in libs]
        condition_str = ", ".join(["'%s'" % cond for cond in conditions])
        if not self._fc_shrinkage_off:
            beta_prior_str = ", betaPrior=TRUE"
        elif self._fc_shrinkage_off:
            beta_prior_str = ""
        file_content = self._deseq_script_template() % (
            self._gene_wise_quanti_combined_path, self._first_data_column-1,
            len(libs), self._first_data_column, libs_str, condition_str,
            beta_prior_str,
            self._deseq_pca_heatmap_path)
        file_content += self._comparison_call_strings(conditions)
        deseq_fh = open(self._deseq_script_path, "w")
        deseq_fh.write(file_content)
        deseq_fh.close()

    def run_deseq(self):
        call(["Rscript", self._deseq_script_path])

    def merge_counting_files_with_results(self):
        for comparison_file, combo in self._comparison_files_and_combos:
            output_fh = open("%s/%s" % (
                self._deseq_extended_folder,
                comparison_file.replace(
                    ".csv", "_with_annotation_and_countings.csv")), "w")
            output_fh.write("# Reference library (divisor): {}\n".format(
                combo[1]))
            output_fh.write("# Comparison library (numerator): {}\n".format(
                combo[0]))
            try:
                deseq_result_fh = open("%s/%s" % (
                    self._deseq_raw_folder, comparison_file))
            except:
                sys.stderr.write("Apparently DESeq did not generate the "
                                 "file \"%s\". Extension stopped.\n" %
                                 comparison_file)
                continue
            for counting_file_row, comparison_file_row in zip(
                csv.reader(open(
                    self._gene_wise_quanti_combined_path), delimiter="\t"),
                    csv.reader(deseq_result_fh, delimiter="\t")):
                if comparison_file_row[0] == "baseMean":
                    # Add another column to the header
                    comparison_file_row = [""] + comparison_file_row
                    # Extend column description
                    counting_file_row[self._first_data_column:] = [
                        "%s raw countings" % lib_name for lib_name in
                        counting_file_row[self._first_data_column:]]
                output_fh.write("\t".join(
                    counting_file_row + comparison_file_row[1:]) + "\n")
            output_fh.close()

    def _condition_combos(self, conditions):
        non_redundant_conditions = set(conditions)
        for cond1 in non_redundant_conditions:
            for cond2 in non_redundant_conditions:
                if not cond1 == cond2:
                    yield((cond1, cond2))

    def _comparison_call_strings(self, conditions):
        call_string = ""
        condition_combos = self._condition_combos(conditions)
        self._comparison_files_and_combos = []
        cooks_cutoff_str = ""
        if self._cooks_cutoff_off:
            cooks_cutoff_str = ", cooksCutoff=FALSE"
        for index, condition_combo in enumerate(condition_combos):
            call_string += ("comp%s <- results(dds, contrast="
                            "c('condition','%s', '%s')%s)\n" % (
                                index, condition_combo[0], condition_combo[1],
                                cooks_cutoff_str))
            comparison_file = "deseq_comp_%s_vs_%s.csv" % (
                condition_combo[0], condition_combo[1])
            self._comparison_files_and_combos.append(
                (comparison_file, list(condition_combo)))
            call_string += (
                "write.table(comp%s, file='%s/%s', "
                "quote=FALSE, sep='\\t')\n" % (
                    index, self._deseq_raw_folder, comparison_file))
        return call_string

    def _deseq_script_template(self):
        return (
            "library('DESeq2')\n"
            "rawCountTable <- read.table('%s', skip=1, sep='\\t', "
            "quote='', comment.char='', "
            "colClasses=c(rep('character',%s), rep('numeric',%s)))\n"
            "countTable <- round(rawCountTable[,%s:length(names("
            "rawCountTable))])\n"
            "libs <- c(%s)\n"
            "conds <- c(%s)\n"
            "colnames(countTable) <- libs\n"
            "samples <- data.frame(row.names=libs, condition=conds, "
            "lib=libs)\n"
            "dds <- DESeqDataSetFromMatrix(countData=countTable, "
            "colData=samples, design=~condition)\n"
            "dds <- DESeq(dds%s)\n\n"
            "# PCA plot\n"
            "pdf('%s')\n"
            "rld <- rlog(dds)\n"
            "print(plotPCA(rld, intgroup=c('condition')))\n\n"
            "# Heatmap\n"
            "library('RColorBrewer')\n"
            "library('gplots')\n"
            "distsRL <- dist(t(assay(rld)))\n"
            "mat <- as.matrix(distsRL)\n"
            "rownames(mat) <- with(colData(dds), "
            "paste(lib, sep=' : '))\n"
            "hmcol <- colorRampPalette(brewer.pal(9, 'GnBu'))(100)\n"
            "heatmap.2(mat, trace='none', col = rev(hmcol), "
            "margin=c(13, 13))\n")
